from abc import abstractmethod
from typing import Type

from elliptic.Kernel.Context import ContextDelegate
from elliptic.Kernel.Contract import DSLContract, DSLImplementation
from elliptic.Kernel.Expression import Expression


class MeshQLFunction:
    """MeshQL function. If builtin is set to True, should be an internal function.
    """

    def __init__(self, fun, builtin=False):
        self.fun = fun
        self.builtin = builtin


class ComputerImplementationBase(DSLImplementation):

    @abstractmethod
    def map_delegate(self, fun: MeshQLFunction) -> Type[ContextDelegate]:
        raise NotImplementedError

    @abstractmethod
    def reduce_delegate(self, fun: MeshQLFunction) -> Type[ContextDelegate]:
        raise NotImplementedError


class ComputerContract(DSLContract[ComputerImplementationBase]):

    def Map(self, fun: MeshQLFunction) -> 'ComputerContract':
        return self.append_tree(Expression(self.dsl_impl.map_delegate(fun), "Map"))

    def Reduce(self, fun: MeshQLFunction) -> 'ComputerContract':
        return self.append_tree(Expression(self.dsl_impl.reduce_delegate(fun), "Reduce"))



from abc import abstractmethod
from typing import Type

from elliptic.Kernel.Context import ContextDelegate
from elliptic.Kernel.Contract import DSLContract, DSLImplementation
from elliptic.Kernel.Expression import Expression


class ManagerImplementationBase(DSLImplementation):

    @abstractmethod
    def solve_delegate(self) -> Type[ContextDelegate]:
        raise NotImplementedError

    @abstractmethod
    def store_delegate(self) -> Type[ContextDelegate]:
        raise NotImplementedError


class ManagerContract(DSLContract[ManagerImplementationBase]):

    def Solve(self) -> 'ManagerContract':
        return self.append_tree(Expression(self.dsl_impl.solve_delegate(), "Solve"))

    def Store(self) -> 'ManagerContract':
        return self.append_tree(Expression(self.dsl_impl.store_delegate(), "Store"))



from .Computer import ComputerContract, ComputerImplementationBase
from .Manager import ManagerContract, ManagerImplementationBase
from .Selector import SelectorContract, SelectorImplementationBase


class MeshQLContract(ComputerContract, ManagerContract, SelectorContract):
    pass



from typing import Type

from elliptic.Kernel.Context import ContextDelegate

from elliptic_meshql.Computer import ComputerImplementationBase
from elliptic_meshql.Manager import ManagerImplementationBase
from elliptic_meshql.MeshQL import MeshQLContract
from elliptic_meshql.Selector import SelectorImplementationBase


class MeshQLImplementationBase(ComputerImplementationBase, ManagerImplementationBase, SelectorImplementationBase):
    def where_delegate(self, conditions):
        pass

    def map_delegate(self, fun) -> Type[ContextDelegate]:
        pass

    def reduce_delegate(self, fun) -> Type[ContextDelegate]:
        pass

    def solve_delegate(self) -> Type[ContextDelegate]:
        pass

    def store_delegate(self) -> Type[ContextDelegate]:
        pass

    def by_ent_delegate(self, dim: int) -> Type[ContextDelegate]:
        pass

    def by_adj_delegate(self, bridge_dim: int, to_dim: int) -> Type[ContextDelegate]:
        pass

    def base_delegate(self) -> Type[ContextDelegate]:
        pass


def test_meshql_interface():
    root = MeshQLContract(MeshQLImplementationBase())

    root.Base().ByEnt(3).ByAdj(2, 3).Where(cond=True).Map(None).Reduce(None).Solve().Store()



from abc import abstractmethod
from typing import Type, cast

from elliptic.Kernel.Context import ContextDelegate
from elliptic.Kernel.Contract import DSLContract, DSLImplementation
from elliptic.Kernel.Expression import Expression


class SelectorImplementationBase(DSLImplementation):

    @abstractmethod
    def by_ent_delegate(self, dim: int) -> Type[ContextDelegate]:
        raise NotImplementedError

    @abstractmethod
    def by_adj_delegate(self, bridge_dim: int, to_dim: int) -> Type[ContextDelegate]:
        raise NotImplementedError

    @abstractmethod
    def where_delegate(self, conditions):
        raise NotImplementedError


class SelectorContract(DSLContract[SelectorImplementationBase]):

    def ByEnt(self, dim: int) -> 'SelectorContract':
        return self.append_tree(Expression(self.dsl_impl.by_ent_delegate(dim), "ByEnt"))

    def ByAdj(self, bridge_dim: int, to_dim: int) -> 'SelectorContract':
        return self.append_tree(Expression(self.dsl_impl.by_adj_delegate(bridge_dim, to_dim), "ByAdj"))

    def Where(self, **conditions):
        return self.append_tree(Expression(self.dsl_impl.where_delegate(conditions), "Where"))
